﻿Telepathy = LibStub("AceAddon-3.0"):NewAddon("Telepathy", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0", "AceSerializer-3.0" );
local L = LibStub("AceLocale-3.0"):GetLocale("Telepathy", false)

--local function cache
local Time = Time
local ipairs = ipairs
local pairs = pairs
local tinsert = tinsert
local _

--debug tracing switch
local teleDebug = false

--
local ButtonCache = 
	{ 
		Foreign = {}, 
		Local = {},
		RequestFrames = {}
	}
--[[ LinkedMembers is the table of party/raid members who have published their CDs
	Entry:
		{
			Name,
			Abilities = { {info, hidden} },
			PanelButtons = { buttonrefs },
			Panel = wrappanel
		}
]]
local LinkedMembers = {} 

--[[ PublishedAbilities is the table of your abilities/items that you have published, containing
	details about the ability as well as a reference to the actual ability button
]]	
local PublishedAbilities = {}

-- button dimensions
local borderPadding = 10

-- opacity settings
local opacityCD = 0.7
local opacityAvailable = 1

-- comm prefixes
local comPublish = "TelepathyPub"
local comCD = "TelepathyCD"
local comRequest = "TelepathyReq"
local comRefresh = "TelepathyRefresh"

-- request sound files
local sounds = {
	["--None--"] = "",
	["Wisp"] = "Sound\\Event Sounds\\Wisp\\WispYes1.wav",
	["Ding"] = "Sound\\interface\\AuctionWindowOpen.wav",
	["Magic Wand"] = "Sound\\Item\\UseSounds\\iMagicWand1.wav",
	["The Sofie Whip"] = "Sound\\Item\\Weapons\\Whip\\BullWhipHit1.wav",
	["PVP Flag Return"] = "Sound\\Spells\\PVPFlagReturned.wav",
	["PVP Enter Queue"] = "Sound\\Spells\\PVPEnterQueue.wav",
	["Bell"] = "Sound\\Spells\\ShaysBell.wav",
	["Spell Break"] = "Sound\\Spells\\SpellBreakTarget.wav",
	["Throw Net"] = "Sound\\Spells\\ThrowNet.wav",
	["Bonk 1"] = "Sound\\Spells\\Bonk1.wav",
	["Bonk 2"] = "Sound\\Spells\\Bonk2.wav",
	["Bonk 3"] = "Sound\\Spells\\Bonk3.wav",
	["Illidan Blades"] = "Sound\\Spells\\IllidanGetBlades.wav",
	["Rein Deer"] = "Sound\\Spells\\SummonReindeer.wav"
	}

-- current comm channel (to change when party/raid changes)
local currentChannel = "PARTY"

-- the spell book for chosing which spells you are interested in
local Telepathy_SpellBook = nil

--colors for player name labels
local Colors = {
		["Class"]={
					["HUNTER"] = { 0.67, 0.83, 0.45, 1 },
					["WARLOCK"] = { 0.58, 0.51, 0.79, 1 },
					["PRIEST"] = { 1.0, 1.0, 1.0, 1 },
					["PALADIN"] = { 0.96, 0.55, 0.73, 1 },
					["MAGE"] = { 0.41, 0.8, 0.94, 1 },
					["ROGUE"] = { 1.0, 0.96, 0.41, 1 },
					["DRUID"] = { 1.0, 0.49, 0.04, 1 },
					["SHAMAN"] = { 0.14, 0.35, 1.0, 1 },
					["WARRIOR"] = { 0.78, 0.61, 0.43, 1 },
					["DEATHKNIGHT"] = { 0.77, 0.12, 0.23, 1 },
					["MONK"] = { 0.33, 0.54, 0.52, 1 },
					["UNKNOWN"] = { 0.77, 0.77, 0.77, 1 },
					}
}

--options
local options = {
    name = "Telepathy",
    handler = Telepathy,
    type = "group",
    args = {
		SpellBook = {
			order = 0,
			name = L["Spell Book"],
			type = "execute",
			desc = L["Open/Close the Telepathy Spellbook"],
			func = function() Telepathy:CreateSpellBook() end,
		},
		Layout = {
			name = L["Layout"],
			type = "group",
			guiInline = true,
			order = 1,
			args = {
				help = {
					order = 0,
					type = "description",
					name = L["Modify the button size and padding options for the combat window"],
				},
				mybuttonsize = {
					order = 1,
					type = "range",			
					min = 16,
					max = 54,
					step = 1,
					bigStep = 1,
					name = L["My Button Size"],
					desc = L["The size of your own ability buttons in the combat window and spell book"],
					get = "GetButtonSizeLocal",
					set = "SetButtonSizeLocal",
					width = "double"
				},
				foreignbuttonsize = {
					order = 2,
					type = "range",			
					min = 16,
					max = 54,
					step = 1,
					bigStep = 1,
					name = L["Foreign Button Size"],
					desc = L["The size of foreign ability buttons in the combat window and spell book"],
					get = "GetButtonSizeForeign",
					set = "SetButtonSizeForeign",
					width = "double"
				},
				buttonpadding = {
					type = "range",
					min = 0,
					max = 20,
					step = 1,
					bigStep = 1,
					name = L["Button Padding"],
					desc = L["The padding space between ability buttons in the combat window and spell book"],
					get = "GetButtonPadding",
					set = "SetButtonPadding",
					width = "double"
				},
			}		
		},
		Opacity = {
			name = L["Opacity"],
			type = "group",
			guiInline = true,
			order = 2,
			args = {
				help = {
					order = 0,
					type = "description",
					name = L["Opacity affects the transparency of the combat window buttons, depending on whether or not the cooldown is available"],
				},
				oncdopacity = {
					order = 1,
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					bigStep = 0.05,
					name = L["On Cooldown Opacity"],
					desc = L["The opacity level for abilities which are on cooldown"],
					get = "GetOpacityCD",
					set = "SetOpacityCD",			
					width = "double"
				},
				offcdopacity = {
					order = 2,
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					bigStep = 0.05,
					name = L["Available Opacity"],
					desc = L["The opacity level for abilities which are available"],
					get = "GetOpacityAvailable",
					set = "SetOpacityAvailable",			
					width = "double"
				},
			}
		},
		RequestSound = {
			name = L["Request Sound"],
			type = "group",
			guiInline = true,
			order = 3,
			args = {
				help = {
					order = 0,
					type = "description",
					name = L["Select a sound to play when one of your abilities is requested by another player"],
				},
				Sound = {
                    order = 1,
                    type = "select",
                    width = 'normal',
                    name = L["Sound"],
                    desc = L["Sound file to play when one of your abilities is requested by another player"],
                    values = function()
							local t = {}							
							for k,v in pairs(sounds) do t[k] = k; end
							sort(t);
							return t;
					end,
					get = "GetRequestSound",
					set = "SetRequestSound"
                },
			}
		},
		nametrim = {
            type = "range",
			min = 0,
			max = 8,
			step = 1,
			bigStep = 1,
            name = L["Name Trim Length"],
            desc = L["The abbreviated player name label length to show for other player's abilities. Set to ZERO for no trimming. Changes to this may require a interface reload."],
            get = "GetNameTrimLength",
            set = "SetNameTrimLength",			
			width = "double"
        },
		hidesolo = {
			type = "toggle",		
			name = L["Hide While Solo"],
            desc = L["Turn this on to have the Telepathy frames hidden while you are not in a party or raid."],
            get = "GetHideSolo",
            set = "SetHideSolo",
			width = "double"
		},
		dockpanels = {
			type = "toggle",		
			name = L["Dock Panels"],
            desc = L["Docks the foreign and local action button panels together. Turn off to move each panel individually."],
            get = "GetDockPanels",
            set = "SetDockPanels",
			width = "double"
		},	
		lockcombat = {
			type = "toggle",		
			name = L["Combat Lock"],
            desc = L["Lock the Telepathy frames when combat starts."],
            get = "GetCombatLock",
            set = "SetCombatLock",
			width = "double"
		},	
    }
}
local defaults = {
			char = {
				ButtonSizeLocal = 28,
				ButtonSizeForeign = 28,
				ButtonPadding = 4,
				--main panel sizing
				MainFrameWidth = 200,
				MainFrameHeight = 400,
				MainFrameX = 0,
				MainFrameY = 0,
				--
				NameTrimLength = 3,
				OpacityCD = 0.8,
				OpacityAvailable = 1,
				RequestSound = "Ding",
				HideSolo = false,
				DockPanels = true,
				--if local and foreign panels are undocked, then these are the coordinates of the local panel
				LocalX = 0,
				LocalY = 0,
				LocalWidth = 200,
				LocalHeight = 80,
				--
				LockInCombat = true,
				Published = {}, -- published abilities
				Subscribed = {} -- foreign abilities you have subscribed to
			}
}

-- profile db pointer
local db


-------------------------------------
-- frame resize code
local function makeSizable(frame, frameleveloffset)
	local grip = CreateFrame("Frame", nil, frame)
	grip:EnableMouse(true)
	grip:SetFrameLevel(frame:GetFrameLevel() + (frameleveloffset or 0))
	local tex = grip:CreateTexture()
	grip.tex = tex
	tex:SetTexture([[Interface\BUTTONS\UI-AutoCastableOverlay]])
	tex:SetTexCoord(0.619, 0.760, 0.612, 0.762)
	tex:SetDesaturated(true)
	tex:ClearAllPoints()
	tex:SetPoint("TOPLEFT")
	tex:SetPoint("BOTTOMRIGHT", grip, "TOPLEFT", 19, -18)
	grip:SetWidth(22)
	grip:SetHeight(21)
	grip:SetScript("OnMouseDown", function(self)
		self:GetParent():StartSizing()
	end)
	grip:SetScript("OnMouseUp", function(self)
		self:GetParent():StopMovingOrSizing()
	end)

	grip:ClearAllPoints()
	grip:SetPoint("BOTTOMRIGHT")
	grip:SetScript("OnEnter", function(self)
		self.tex:SetDesaturated(false)
	end)
	grip:SetScript("OnLeave", function(self)
		self.tex:SetDesaturated(true)
	end)
	frame.grip = grip
end

local function makeUnsizable(frame)
	local grip = frame.grip
	if grip then
		grip:Hide()
		grip:SetParent(nil)
		grip = nil
	end
end
-------------------------------------

-------------------------------------
-------------------------------------
-- Settings handlers
function Telepathy:GetButtonSizeLocal(info)
	return db.ButtonSizeLocal;
end
function Telepathy:SetButtonSizeLocal(info, newval)
	db.ButtonSizeLocal = newval	
	for k,v in pairs(PublishedAbilities) do
		if v.button then
			v.button:SetHeight(newval)
			v.button:SetWidth(newval)
		end
	end	
	if self.MainPanel then
		self.MainPanel:Refresh()
	end
end
--
function Telepathy:GetButtonSizeForeign(info)
	return db.ButtonSizeForeign;
end
function Telepathy:SetButtonSizeForeign(info, newval)
	db.ButtonSizeForeign = newval
	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button then
				info.button:SetHeight(newval)
				info.button:SetWidth(newval)
				info.container:SetHeight(newval+14)
				info.container:SetWidth(newval)
			end
		end
	end
	if self.ForeignPanel then
		self.ForeignPanel:Refresh()
	end
end
--
function Telepathy:GetButtonPadding(info)
	return db.ButtonPadding;
end
function Telepathy:SetButtonPadding(info, newval)
	db.ButtonPadding = newval	
	if self.MainPanel then
		self.MainPanel:SetPadding(newval)
		self.MainPanel:Refresh()
	end
	if self.ForeignPanel then
		self.ForeignPanel:SetPadding(newval)
		self.ForeignPanel:Refresh()
	end
end
--
function Telepathy:GetNameTrimLength(info)
	return db.NameTrimLength
end
function Telepathy:SetNameTrimLength(info, newval)
	db.NameTrimLength = newval
end
--
function Telepathy:GetOpacityCD(info)
	return db.OpacityCD
end
function Telepathy:SetOpacityCD(info, newval)
	db.OpacityCD = newval
	opacityCD = newval

	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button and info.cd then
				if info.cd:IsShown() then
					info.container:SetAlpha(opacityCD)
					--info.cd:SetAlpha(opacityCD)
					--info.button:SetAlpha(opacityCD)
				end
			end
		end
	end
end
--
function Telepathy:GetOpacityAvailable(info)
	return db.OpacityAvailable
end
function Telepathy:SetOpacityAvailable(info, newval)
	db.OpacityAvailable = newval
	opacityAvailable = newval

	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button and info.cd then
				if not info.cd:IsShown() then
					info.container:SetAlpha(opacityAvailable)
					--info.cd:SetAlpha(opacityAvailable)
					--info.button:SetAlpha(opacityAvailable)
				end
			end
		end
	end
end
--
function Telepathy:GetRequestSound(info)
	return db.RequestSound
end
function Telepathy:SetRequestSound(info, newval)
	db.RequestSound = newval
	self:Trace("Setting request sound to", newval)

	self:Trace(info.arg)
	local sound_file = sounds[newval]	
	if sound_file and strlen(sound_file) > 0 then
		self:Trace("Playing soundfile", sound_file)
		PlaySoundFile(sound_file)		
	end
end
--
function Telepathy:GetHideSolo(info)
	return db.HideSolo
end
function Telepathy:SetHideSolo(info, newval)
	db.HideSolo = newval
	self:CheckHideSolo()
end
--
function Telepathy:GetDockPanels(info)
	return db.DockPanels
end
function Telepathy:SetDockPanels(info, newval)
	db.DockPanels = newval
	local mp, mf = self.MainPanel, self.MainFrame
	mp:ClearAllPoints()
		
	if newval then
		--docking the local panel to the main frame
		mp:SetParent(mf)
		mp:SetPoint("BOTTOMLEFT")
		makeUnsizable(mp)
	else
		--undocking the local panel from the main frame
		mp:SetParent(UIParent)
		mp:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db.LocalX, db.LocalY)
		makeSizable(mp, 5)		
		mp:SetMinResize(db.ButtonSizeLocal,db.ButtonSizeLocal)
	end

	mp:SetResizable(true)
	mp:SetMovable(true)

	self:InitialResizeWorkaround()
	self:StoreLocalPanelPosition()	
end
--
function Telepathy:GetCombatLock(info)
	return db.CombatLock
end
function Telepathy:SetCombatLock(info, newval)
	db.CombatLock = newval
end
--/Settings Handlers
-------------------------------------
-------------------------------------


function Telepathy:Trace(...)
	if teleDebug then
		self:Print(...)
	end
end

function Telepathy:OnInitialize()
    -- Called when the addon is loaded
    -- Print a message to the chat frame
    self:Trace("OnInitialize Event Fired: Hello")
end

local function sortContainerByClass(a, b)
	local r = true
	if a and b then
		local aMT, bMT, aI, bI = a.MemberTable, b.MemberTable, a.info, b.info
		if aMT and bMT then
			if aMT == bMT then
				r = aI.id < bI.id
			else
				if aMT.ClassName ~= bMT.ClassName then
					r = aMT.ClassName < bMT.ClassName
				elseif aI.id == bI.id then
					r = aMT.Name < bMT.Name
				else
					r = aI.id < bI.id
				end
			end
		elseif not aMT then
			r = true
		elseif not bMT then
			r = false
		end
	elseif not a then
		r = true		
	elseif not b then
		r = false
	end
	return r
end

local function SortButtonsByClassPlayer(t)
	sort(t, sortContainerByClass)

	--apply wrap hints based on class/ability
	local oldclass = nil
	for i, v in ipairs(t) do
		v.WrapHint = nil
		if v and v.MemberTable then
			if oldclass ~= v.MemberTable.ClassName then
				v.WrapHint = "NEWLINE"
				oldclass = v.MemberTable.ClassName
			end
		end
	end
end

local function GetCurrentActiveSpecName(specificSpecNumber)
	local getnumber = 1
	if ( specificSpecNumber ) then 
		Telepathy:Trace("Getting active spec name for spec number", getnumber)
		getnumber = GetSpecialization(false, false, getnumber)
	else
		getnumber = GetSpecialization()
	end
	local _, specname = GetSpecializationInfo(getnumber)
	return specname
end

local function GetCurrentChannel()
	local i = GetNumGroupMembers()
	if i > 0 then
		if IsInRaid() == true then
			currentChannel = "RAID"
		end
	else
		i = GetNumGroupMembers()
		if i > 0 then
			currentChannel = "PARTY"
		end		
	end
	return currentChannel, i
end

local function ClassColor(name)
	local class, classname = nil, "UNKNOWN"
	if UnitExists(name) then		
		class, classname = UnitClass(name)
		if not classname then
			classname = "UNKNOWN" --should really be SHAMAN, cos shaman are filled with win
		end
	end
	if teleDebug then
		classname = LinkedMembers[name].ClassName
	end
	local c = Colors["Class"][classname]
	return c[1], c[2], c[3]
end

local function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function Telepathy.OnMainFrameSizeChanged(frame, w, h)
	--store in profile db
	db.MainFrameWidth = w
	db.MainFrameHeight = h

	local mp = Telepathy.MainPanel
	local fp = Telepathy.ForeignPanel
	local width = w
	local heightmp = mp:GetHeight() -- this one auto sizes, so take it's height into account
	local heightfp = h
	if db.DockPanels then
		heightfp = heightfp - heightmp
		mp:SetWidth(width)		
		mp:Refresh()
	end
	if fp then
		fp:SetWidth(width)
		fp:SetHeight(heightfp)
		fp:Refresh()	
	end
end

function Telepathy:StoreMainFramePosition()
	db.MainFrameX = self.MainFrame:GetLeft()
	db.MainFrameY = self.MainFrame:GetTop()
	if db.DockPanels then
		self:StoreLocalPanelPosition()
	end
end

function Telepathy:StoreLocalPanelPosition()
	local mp = self.MainPanel	
	db.LocalX = mp:GetLeft()
	db.LocalY = mp:GetTop()
	db.LocalWidth = mp:GetWidth()
	db.LocalHeight = mp:GetHeight()
end

function Telepathy:CreateMainFrame()

	local f = CreateFrame("Frame", "TelepathyMainFrame", UIParent)
	f:ClearAllPoints()
	if db.MainFrameX == 0 and db.MainFrameY == 0 then
		f:SetPoint("CENTER",UIParent)
	else
		f:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT",db.MainFrameX,db.MainFrameY)
	end
	local buttonSize = db.ButtonSizeLocal
	f:EnableMouse(true)	
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(buttonSize+db.ButtonPadding,buttonSize+db.ButtonPadding)
	f:SetWidth(db.MainFrameWidth)
	f:SetHeight(db.MainFrameHeight)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart",function() if (not InCombatLockdown() or not db.CombatLock) then f.StartMoving(); end; end)
	f:SetScript("OnDragStop",function() f:StopMovingOrSizing(); Telepathy:StoreMainFramePosition(); end)
	f:SetScript("OnSizeChanged", Telepathy.OnMainFrameSizeChanged)
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()

	makeSizable(f, 2)

	self.MainFrame = f
end

function Telepathy:CreateMainPanel()
	local parent = self.MainFrame
	if not db.DockPanels then parent = UIParent end
	local f = CreateWrapPanel(parent)
	f:SetScript("OnMouseDown", function(frame, button)
			  if button == "LeftButton" then
					local parent = frame:GetParent()
					if (not InCombatLockdown() or not db.CombatLock) then 
						if db.DockPanels and not parent.isMoving then
							parent:StartMoving()
							parent.isMoving = true									
						elseif not db.DockPanels and not frame.isMoving then
							frame:StartMoving()
							frame.isMoving = true	
						end
					end
				end
			end)
	f:SetScript("OnMouseUp", function(frame, button)
			  if button == "LeftButton" then
				  local parent = frame:GetParent()
				  if db.DockPanels and parent.isMoving then
					parent:StopMovingOrSizing();
					Telepathy:StoreMainFramePosition();
					parent.isMoving = false;
				  elseif not db.DockPanels and frame.isMoving then
				    frame:StopMovingOrSizing();
					Telepathy:StoreLocalPanelPosition();
					frame.isMoving = false;
				  end
			  end
	end)
	f:SetScript("OnSizeChanged", function(frame) frame:Refresh(); Telepathy:StoreLocalPanelPosition(); end)
	local buttonSize = db.ButtonSizeLocal
	f.Tag = "MainPanel"
	f:SetPadding(db.ButtonPadding)
	f:SetMargins(5)
	f:ClearAllPoints()
	f:SetResizable(true)
	f:SetMovable(true)
	if db.DockPanels then
		f:SetPoint("BOTTOMLEFT",self.MainFrame)
		f:SetHeight(buttonSize+buttonPadding)	
	else
		if db.LocalX == 0 and db.LocalY == 0 then
			f:SetPoint("CENTER")
		else
			f:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", db.LocalX, db.LocalY)
		end
		f:SetWidth(db.LocalWidth)		
		f:SetHeight(db.LocalHeight)		
		makeSizable(f,5)
		f:SetMinResize(db.ButtonSizeLocal,db.ButtonSizeLocal)
	end
	f:EnableMouse(true)		
	
	f.MinHeight = buttonSize+buttonPadding
	f.AutoSizeY = true	
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()
	self.MainPanel = f
	f:SetContext(ButtonCache.Local)
end

function Telepathy:CreateForeignPanel()
	local parent = self.MainFrame
	local f = CreateWrapPanel(parent)
	f:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not parent.isMoving then
				  if (not InCombatLockdown() or not db.CombatLock) then
						parent:StartMoving();
						parent.isMoving = true;
				  end
				end
			end)
	f:SetScript("OnMouseUp", function(self, button)
			  if button == "LeftButton" and parent.isMoving then
			   parent:StopMovingOrSizing();
			   Telepathy:StoreMainFramePosition();
			   parent.isMoving = false;
			  end
			end)
	f.Tag = "ForeignPanel"
	f:SetPadding(db.ButtonPadding)
	f:SetMargins(5)
	f:ClearAllPoints()
	f:SetPoint("TOPLEFT",parent,"TOPLEFT",0,0)
	f:EnableMouse(true)
	--[[
	f:SetScript("OnUpdate", function(frame, elapsed)
		frame.UpdateDelay = elapsed + (frame.UpdateDelay or 0)
		if frame.UpdateDelay > 0.5 then
			local updateTable = frame.MemberTablesForUpdate
			if updateTable and #updateTable > 0 then
				local t = updateTable[#updateTable]
				Telepathy:RefreshForeignPanel(t)
				updateTable[#updateTable] = nil
				if #updateTable == 0 then
					frame.UpdateDelay = 0 --no more processing
				end
			end			
		end
	end)
	f.MarkTableForUpdate = function(frame, t)
		local MemberTablesForUpdate = frame.MemberTablesForUpdate or {}
		frame.MemberTablesForUpdate = MemberTablesForUpdate
		local bNeedsUpdate = true
		for k,v in pairs(MemberTablesForUpdate) do
			if v == t then
				bNeedsUpdate = false
				break
			end
		end
		if bNeedsUpdate then
			tinsert(MemberTablesForUpdate, t)
		end
	end
	]]
	local buttonSize = db.ButtonSizeForeign
	local buttonPadding = db.ButtonPadding
	f:SetHeight((buttonSize+buttonPadding)*4+borderPadding)
	f.MinHeight = buttonSize+buttonPadding+borderPadding	
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()
	f:SetContext(ButtonCache.Foreign)
	self.ForeignPanel = f
end

local function AttachTooltips(frame, info)
	if info.type == "spell" then
		frame:HookScript("OnEnter", function(self) GameTooltip:SetOwner(self); GameTooltip:SetSpellByID(info.id); end )
	elseif itemid then
		frame:HookScript("OnEnter", function(self) GameTooltip:SetOwner(self); GameTooltip:SetItemByID(info.id); end )
	end
	frame:HookScript("OnLeave", function() GameTooltip:Hide(); end )
end

function Telepathy:CreateSpellBook()
	
	if Telepathy_SpellBook then
		if Telepathy_SpellBook:IsVisible() then
			Telepathy_SpellBook:Hide()
			Telepathy_SpellBook = nil --destroy
			return
		end		
	end	

	local buttonSize = db.ButtonSizeForeign
	local wrap
	local f = CreateFrame("Frame", "TelepathySpellBook", UIParent)
	Telepathy_SpellBook = f
	f:SetFrameLevel(Telepathy.MainFrame:GetFrameLevel() + 5)
	f:ClearAllPoints()
	f:SetPoint("CENTER",UIParent)
	f:EnableMouse(true)	
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(max(6*buttonSize,500),8*buttonSize)
	f:SetWidth(500)
	f:SetHeight(400)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart",f.StartMoving)
	f:SetScript("OnDragStop",f.StopMovingOrSizing)	
	f:SetBackdrop({
		bgFile = [[Interface\AchievementFrame\UI-Achievement-Parchment-Horizontal]], tile = false, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)

	f.TitleBg = f:CreateTexture(nil,"ARTWORK")
	f.TitleBg:SetPoint("BOTTOM", f, "TOP", 0, -32)
	f.TitleBg:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
	f.TitleBg:SetHeight(59)
	f.TitleBg:SetWidth(f:GetWidth())
			
	f.Title = f:CreateFontString(nil,"ARTWORK","GameFontNormal")
	f.Title:SetPoint("BOTTOM", f.TitleBg, "BOTTOM", 0, 34)
	f.Title:SetText("Telepathy Spellbook")

	f:Show()
	makeSizable(f, 5)

	--close button
	local closebutton = CreateFrame("Button", nil, f, "UIPanelButtonTemplate")
	closebutton:ClearAllPoints()	
	closebutton:SetHeight(24)
	closebutton:SetWidth(70)
	closebutton:SetText(L["Close"])	
	closebutton:Show()
	closebutton:SetScript("OnClick", function(b) Telepathy:CreateSpellBook() end)
	--close button position set relative to the lower wrap panel below

	--refresh button
	local refreshbutton = CreateFrame("Button", nil, f, "UIPanelButtonTemplate")
	refreshbutton:ClearAllPoints()	
	refreshbutton:SetHeight(24)
	refreshbutton:SetWidth(90)
	refreshbutton:SetText(L["Refresh"])	
	refreshbutton:Show()
	refreshbutton:SetScript("OnClick", function(b) Telepathy:RequestBigRefresh() end)
	refreshbutton:SetPoint("RIGHT", closebutton, "LEFT", -6, 0)
	--
	--hint label - center of spellbook
	local hint1 = f:CreateFontString(nil,"ARTWORK")
	hint1:SetFont("Fonts\\FRIZQT__.TTF",11) --(GameFontNormal:GetFont())
	hint1:SetText(L["|cffCED000Left-click a player's abilitiy to add it to the Combat Window. Right-click to remove.|r"])
	hint1:SetPoint("CENTER")
	hint1:SetWidth(120)
	hint1:Show()

	--foreign wrap panel
	wrap = CreateWrapPanel(f)
	wrap:ClearAllPoints()
	wrap:SetPoint("TOPLEFT")
	wrap:SetPoint("TOPRIGHT")
	wrap:EnableMouse(true)
	wrap:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not f.isMoving then
			   f:StartMoving();
			   f.isMoving = true;
			  end
			end)
	wrap:SetScript("OnMouseUp", function(self, button)
			  if button == "LeftButton" and f.isMoving then
			   f:StopMovingOrSizing();
			   f.isMoving = false;
			  end
	end)
	wrap.AutoSizeY = true
	wrap:SetMargins(10)
	wrap:SetPadding(5)
	wrap:Show()

	local context = {}
	local button, container

	local f_OnForeignClick = function(b, mousebutton)
					local info = b.info
					Telepathy:Trace("OnClick on spellbook spell:", mousebutton)
					if mousebutton == "LeftButton" then
						info.Hidden = false
						PlaySoundFile("Sound\\interface\\uSpellIconPickup.wav")
					elseif mousebutton == "RightButton" then
						info.Hidden = true
						PlaySoundFile("Sound\\interface\\uSpellIconDrop.wav")
					end
					self:StoreSubscribed(info.MemberTable.Name, info.id, not info.Hidden)		
					Telepathy:RefreshForeignPanel(info.MemberTable)
					--Telepathy.ForeignPanel:MarkTableForUpdate(info.MemberTable)
				end

	f.RepopulateForeignButtons = function()
		local context = wrap.Context or {}
		for k,v in pairs(context) do v:Hide(); v = nil; end
		wipe(context)
		
		for name, membertable in pairs(LinkedMembers) do
			for k,v in pairs(membertable.Abilities) do
				button, _, container = self:CreateForeignAbilityButton(v, wrap, false, membertable.Name)
				self:Trace("Adding container at index",#context + 1,container)
				tinsert(context,container)
				container.MemberTable = membertable
				container.info = v
				button.info = v			
				button:SetScript("OnClick", f_OnForeignClick)
			end
		end
		SortButtonsByClassPlayer(context)
		wrap:SetContext(context)
		wrap:Refresh()
	end


	f.RepopulateForeignButtons()
	
	--create the lower panel for my own spells
	local specPanels = {}
	local p = self:CreateSpellBookSpecPanel(f, GetSpecialization(false, false, 1))
	tinsert(specPanels, p)
	if (  GetNumSpecializations() > 1 ) then
		p = self:CreateSpellBookSpecPanel(f, GetSpecialization(false, false, 2), p)
		tinsert(specPanels, p)
	end
	
	--close button relative to this panel's top right corner
	closebutton:SetPoint("BOTTOMRIGHT", p, "TOPRIGHT", -5, 0)

	f:SetScript("OnSizeChanged", 
		function(self, w, h)
			if wrap:GetWidth() ~= w then
				wrap:SetWidth(w)				
			end
			if wrap:GetHeight() ~= (h - p:GetHeight()) then
				wrap:SetHeight(h - p:GetHeight())
			end
			for k,v in pairs(specPanels) do
				if v:GetWidth() ~= w then
					v:SetWidth(w)
				end
				v.WrapPanel:Refresh()
			end
			wrap:Refresh()
		end
	)
	
end

function Telepathy:CreateSpellBookSpecPanel(spellbookframe, specnumber, bindframe)
	
	local _, specname, _, specicon = GetSpecializationInfo(specnumber)
	self:Trace("Creating spec panel for spec ", specname)

	local f = spellbookframe
	--create the lower panel for my own spells
	local mycontext = {}
	local buttonSize = db.ButtonSizeForeign

	self:Trace("Creating wrap panel")

	local FrameSpecContainer = CreateFrame("Frame", nil, f)
	FrameSpecContainer:ClearAllPoints()
	FrameSpecContainer:SetPoint("BOTTOMLEFT")
	FrameSpecContainer:SetPoint("BOTTOMRIGHT")
	if ( bindframe ) then
		FrameSpecContainer:SetPoint("BOTTOMLEFT", bindframe, "TOPLEFT", 0, -5)
		FrameSpecContainer:SetPoint("BOTTOMRIGHT", bindframe, "TOPRIGHT", 0, -5)
	end
	FrameSpecContainer:SetHeight(max(buttonSize+10,64+10))
	FrameSpecContainer:SetBackdrop({
		--bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		bgFile = [[Interface\AchievementFrame\UI-Achievement-Parchment-Horizontal]], tile = false, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4}
	})	
	FrameSpecContainer:SetBackdropBorderColor(0.2,0.2,0.2)
	FrameSpecContainer:Show()
	
	local spellSlotFrame = FrameSpecContainer:CreateTexture()
	spellSlotFrame:SetTexture("Interface\\Spellbook\\Spellbook-Parts")
	spellSlotFrame:SetTexCoord(0.00390625, 0.27734375, 0.44140625, 0.69531250)
	spellSlotFrame:SetWidth(64)
	spellSlotFrame:SetHeight(64)
	spellSlotFrame:ClearAllPoints()
	spellSlotFrame:SetPoint("LEFT", FrameSpecContainer, "LEFT", 5, 5)
	spellSlotFrame:Show()

	local specTexture = FrameSpecContainer:CreateTexture()
	specTexture:SetTexture(specicon)
	specTexture:SetWidth(34)
	specTexture:SetHeight(34)
	specTexture:ClearAllPoints()
	specTexture:SetPoint("CENTER", spellSlotFrame, "CENTER", -2, 0)
	specTexture:Show()
	
	local p = CreateWrapPanel(FrameSpecContainer)
	p:ClearAllPoints()
	p:SetPoint("BOTTOMLEFT", spellSlotFrame, "BOTTOMRIGHT", 5, 0)
	p:SetPoint("TOPRIGHT", FrameSpecContainer, "TOPRIGHT", 0, 0)
	p:EnableMouse(true)

	self:Trace("Setting script for OnMouseDown")

	p:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not f.isMoving then
				 local infoType, info1, info2 = GetCursorInfo()
				 if not infoType then
					f:StartMoving();
					f.isMoving = true;
				 end
			  end
	end)
	
	self:Trace("Setting script for OnMouseUp")

	p:SetScript("OnMouseUp", function(self, button)
				if button == "LeftButton" then
					local infoType, info1, info2 = GetCursorInfo()
					if not infoType then
						if f.isMoving then
							f:StopMovingOrSizing();
							f.isMoving = false;
						end
					else
						if (infoType and infoType == "spell") then												
							local spell, id = GetSpellBookItemInfo(info1, "spell")
							Telepathy:Trace("Drop on spellbook:", specname, infoType, info1, spell)
							local notcurrentspec = (specname ~= GetCurrentActiveSpecName())
							Telepathy:PublishAbility(specname, id, false, notcurrentspec)
							ClearCursor()
							p.RepopulateButtons()
						else
							Telepathy:Trace("Mouse up on spellbook, cursor info type=", infoType)
						end						
					end
			  end
	end)	
	p.AutoSizeY = true
	p:SetMargins(10)
	p:SetPadding(5)
	p:SetHeight(buttonSize + 10)
	p.MinHeight = p:GetHeight()
	p:SetWidth(f:GetWidth())
	p:SetFrameLevel(f:GetFrameLevel()+5)	

	self:Trace("Showing spec panel for ", specname)

	p:Show()

	--populate context
	p.RepopulateButtons = function()

		local f_OnMyClick = function(b, mousebutton)
			if mousebutton == "RightButton" and b.info then
				local id = b.info.id
				Telepathy:PublishAbility(specname, id, true)
				PublishedAbilities[specname][id] = nil
				PlaySoundFile("Sound\\interface\\uSpellIconDrop.wav")
				p.RepopulateButtons()
			end			
		end

		local context = p.Context or {}
		
		--clear context
		for k,v in pairs(context) do v:Hide() end
		wipe(context)

		if ( PublishedAbilities[specname] ) then

			for k,v in pairs(PublishedAbilities[specname]) do
				local info = v
				local start, duration, enabled
				local b = CreateFrame("Button", nil, p)
				local texture
				if info.type == "spell" then
					local _, _, spellicon = GetSpellInfo(info.id)	
					texture = spellicon
				elseif info.type == "item" then
					local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(info.id)
					texture = itemicon
				end   	 	
				b:SetNormalTexture(texture)
				b:SetHeight(db.ButtonSizeLocal)
				b:SetWidth(db.ButtonSizeLocal)
				b:Show()
				b:SetScript("OnClick", f_OnMyClick)
				b:RegisterForClicks("RightButtonUp")
				b.info = info

				AttachTooltips(b, info)

				tinsert(context, b)
			end
		end
		p:SetContext(context)
		p:Refresh()
	end

	p.RepopulateButtons()
	
	--hint label - center of my wrap panel
	local hint2 = p:CreateFontString(nil,"ARTWORK")
	hint2:SetFont("Fonts\\FRIZQT__.TTF",11) --(GameFontNormal:GetFont())
	hint2:SetText(L["|cffCED000Drag your %s abilities onto this pane to publish them. Right-click to remove.|r"]:format(specname or "Unknown"))
	hint2:ClearAllPoints()
	hint2:SetPoint("TOPRIGHT", p, "TOPRIGHT", -5, 0)
	hint2:SetPoint("BOTTOMRIGHT", p, "BOTTOMRIGHT", -5, 2)
	hint2:SetWidth(120)
	hint2:Show()

	FrameSpecContainer.WrapPanel = p
	return FrameSpecContainer
end

function Telepathy:CreateForeignPlayerPanel(parentframe, membertable, forspellbook)
	if not forspellbook and membertable.Panel then
		return membertable.Panel
	end

	--self:Trace("Creating player panel for", membertable.Name)

	local panel = CreateWrapPanel(parentframe)
	panel.Tag = "TelepathyForeign-"..membertable.Name
	panel:ClearAllPoints()
	panel:SetPoint("TOPLEFT",parentframe)	
	local context = {} -- button list
	if not forspellbook then
		membertable.Panel = panel
		membertable.PanelButtons = context
	end

	--self:Trace("Player panel created")

	panel.AutoSizeY = true
	panel.AutoSizeX = true
	panel.MinHeight = buttonSize	
	panel:SetPadding(4,2)
	panel:SetMargins(5)
	panel.BGTexture=panel:CreateTexture()
	panel.BGTexture:SetAllPoints(panel)
	panel.BGTexture:SetTexture(0.1,0.1,0.1,0.3)

	--create the name of the player here
	--self:Trace("Creating name label")
	local nameLabel = panel:CreateFontString(nil,"ARTWORK")
	nameLabel:SetFont(GameFontNormal:GetFont())
	--self:Trace("Name label created, setting text")
	nameLabel:SetText(membertable.Name)
	local r,g,b =ClassColor(membertable.Name)
	nameLabel:SetTextColor(r,g,b,1)
	nameLabel.WrapYOffset = 3
	panel.MinWidth = nameLabel:GetWidth() + buttonSize + buttonPadding
	self:Trace("Set player panel minwidth=",panel.MinWidth)
	--self:Trace("Inserting name label into context")
	tinsert(context,nameLabel)
	--now all of the published abilities (or just the shown ones, depending on frame type)
	local button, cooldown, container
	local createcooldown = (not forspellbook)
	for k,v in pairs(membertable.Abilities) do
		if forspellbook then
			--self:Trace("Adding foreign ability button")
			button, cooldown, container = self:CreateForeignAbilityButton(v, panel, createcooldown)			
			--self:Trace("Inserting into table")
			tinsert(context,container)
			--self:Trace("Insert done")
			if not forspellbook then
				v.button, v.cd = button, cooldown
			end
			--self:Trace("Moving on...")
		end
	end		
	--self:Trace("Showing panel")
	panel:Show()
	--self:Trace("Setting panel context")
	panel:SetContext(context, true)	
	local prefWidth = panel:PreferredWidth()
	--self:Trace("Getting preferred width returned", prefWidth)
	panel:SetWidth(prefWidth)
	--self:Trace("Refreshing panel")
	panel:Refresh()	

	if not forspellbook then
		tinsert(ButtonCache.Foreign, #ButtonCache.Foreign, panel)
		Telepathy.ForeignPanel:Refresh()
	end
	return panel
end

function Telepathy:CheckHideSolo()
	if db.HideSolo then
		local channel, count = GetCurrentChannel()
		if count > 0 then
			Telepathy.MainFrame:Show()
		else
			Telepathy.MainFrame:Hide()
		end
	else
		Telepathy.MainFrame:Show()
	end
end

local function removeentry(tab, val)
	for k, v in pairs(tab) do
		if v == val then
			tab[k] = nil
		end
	end
end

local function removeientry(tab, val)
	for k, v in pairs(tab) do
		if v == val then
			tremove(tab,k)
		end
	end
end

local function RequestAction(button)
	local info = button.info
	if not info then return end
	if not info.requestable then return end
	if info.cd:IsVisible() then
		Telepathy:Trace("Not requesting action, as it is still on CD!")
		return
	end
	
	PlaySoundFile("Sound\\interface\\iUiInterfaceButtonA.wav")
	Telepathy:ComRequest("a", info.id, info.type, info.Name)
end

function Telepathy:RefreshForeignPanel(membertable)
	
	if not membertable then return end
	--if not membertable.Panel then return end
		
	self:Trace("RefreshForeignPanel(membertable=",membertable.Name,")")	

	--local context = membertable.PanelButtons
	local panel = self.ForeignPanel --membertable.Panel		
	local fcache = ButtonCache.Foreign
	local button, cooldown, container

	for k,v in pairs(membertable.Abilities) do
		if not v.Hidden and not v.button then			
			button, cooldown, container = self:CreateForeignAbilityButton(v, panel, true, membertable.Name)
			self:Trace("Adding container at index",#fcache + 1,container)
			tinsert(fcache,container)
			v.button, v.cd, v.container = button, cooldown, container
			container.MemberTable = membertable
			container.info = v
			self:RefreshSpellCooldown(v)
		elseif v.button and v.Hidden then
			self:Trace("Removing container from foreign cache")
			removeientry(fcache, v.container)
			v.container:Hide()
			v.button = nil --throw the button and cd away
			v.cd = nil
			v.container = nil
		end		
	end
	--local prefwidth = panel:PreferredWidth()
	--self:Trace("Preferred width =",prefwidth)
	--panel:SetWidth(min(prefwidth, self.ForeignPanel:GetWidth()))
	SortButtonsByClassPlayer(fcache)
	panel:Refresh()
end

function Telepathy:AlreadySubscribed(name, id)
	if db.Subscribed[name] and db.Subscribed[name][id] then
		return true
	end
	return false
end

function Telepathy:StoreSubscribed(name, id, hidden)
	local sub = db.Subscribed[name] or {}
	sub[id] = hidden
	db.Subscribed[name] = sub
end

function Telepathy:AddRequester(id, name)
	local spec = GetCurrentActiveSpecName()
	local abilities = PublishedAbilities[spec]
	if not abilities then
		return
	end
	if not abilities[id] then
		return
	end

	ButtonCache.RequestFrames[id] = ButtonCache.RequestFrames[id] or {}
	local requester_table = ButtonCache.RequestFrames[id]
	local label, label_name
	for i,v in ipairs(requester_table) do
		label_name = v.name
		if label_name == name then
			return --already got a request from this person for this item
		end
	end

	--new request from this player for this ability
	local player_table = { name=name, expires=GetTime()+6, id=id }
	tinsert(requester_table, player_table)

	--parent button is the ability button
	local button = abilities[id].button

	--create label
	self:Trace("Creating request label for",id,"from",name)
	label = button:CreateFontString(nil,"ARTWORK")
	label:SetFont("Fonts\\ARIALN.TTF",11, "OUTLINE") --(GameFontNormal:GetFont())
	label:SetText(name)	
	local r,g,b =ClassColor(name)
	label:SetTextColor(r,g,b,1)
	label:SetPoint("TOP", button, "BOTTOM")
	label:Show()
	
	player_table.label = label
	
	--cause the label to remove itself after x seconds
	self:ScheduleTimer("RemoveRequester", 6, player_table)	
end
		
function Telepathy:RemoveRequester(player_table)
	local id = player_table.id
	local request_table = ButtonCache.RequestFrames[id]
	if not request_table then
		return
	end

	if not player_table.label then
		--we may have already thrown this away with a spell cast
		return
	end

	--kill it
	for it,vt in ipairs(request_table) do
		if vt==player_table then
			tremove(request_table, it)
			break
		end
	end
	player_table.label:Hide()
	player_table.label = nil
end

local function OnCDButtonUpdate(self)
	if self.needsupdate then
		self.needsupdate = false
		local i = self.info
		local cd = i.cd

		local start, duration, enabled
		if i.type == "spell" then
			start, duration, enabled = GetSpellCooldown(i.id)
		elseif i.type == "item" then
			start, duration, enabled = GetItemCooldown(i.id)
		end
		cd:SetCooldown(start, duration)

		Telepathy:ComNotifyCD(i.id, i.type, currentChannel, nil)
		if teleDebug then
			Telepathy:ComNotifyCD(i.id, i.type, "WHISPER", UnitName("player"))
		end
	end
end

function Telepathy:RegisterForeignAbility(id, type, requestable, from, cd_remaining, cd_duration, suppress_refresh)
	self:Trace("Registering foreign ability for", from,"with id", id)
	local member = LinkedMembers[from] or { Name=from, Abilities={}}
	LinkedMembers[from] = member
	local info = member.Abilities[id] or { id=id, type=type, requestable=requestable, Name=from, Hidden=true }
	member.Abilities[id] = info
	member.Name = from	
	info.MemberTable = member
	info.cd_duration = cd_duration
	info.cd_start = GetTime() + cd_remaining - cd_duration
	info.Hidden = not self:AlreadySubscribed(from, id)

	if UnitExists(from) then
		_, member.ClassName = UnitClass(from)
	elseif teleDebug then
		--random class
		local i, j = random(1, 10), 1
		for k,v in pairs(Colors["Class"]) do
			if i == j then
				member.ClassName = k
				break
			end
			j = j + 1
		end
	end
	
	if Telepathy_SpellBook and not suppress_refresh then
		Telepathy_SpellBook.RepopulateForeignButtons()
	end
end

function Telepathy:CreateForeignAbilityButton(info, parentframe, includecooldown, playername)
	local id = info.id
	local t = info.type
	local start, duration, enabled	
	local buttonSize = db.ButtonSizeForeign
	
	--create a bottom frame for the player name
	local f = CreateFrame("Frame", nil, parentframe)
	f:SetHeight(buttonSize + 14) --name at the bottom
	f:SetWidth(buttonSize)
	f:Show()
	
	local trimlen = db.NameTrimLength
	if trimlen == 0 then trimlen = 99 end
	local nameforbutton = playername
	if playername and #playername > trimlen then
		nameforbutton = strsub(playername, 1, trimlen)
	end
	
	local nameLabel = f:CreateFontString(nil,"ARTWORK")
	nameLabel:SetFont("Fonts\\ARIALN.TTF",11, "OUTLINE") --(GameFontNormal:GetFont())
	nameLabel:SetText(nameforbutton)
	self:Trace("Set player label text to",nameforbutton)
	local r,g,b =ClassColor(playername)
	nameLabel:SetTextColor(r,g,b,1)
	nameLabel:SetPoint("BOTTOMLEFT", f)
	nameLabel:SetPoint("BOTTOMRIGHT", f)

	nameLabel:Show()
	
	local b = CreateFrame("Button", nil, f)
	b:SetPoint("TOP")
	b.id = id
	b.info = info	
	local texture
	if t == "spell" then
		local _, _, spellicon = GetSpellInfo(id)	
		texture = spellicon
	elseif t == "item" then
		local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(id)
		texture = itemicon
	end   	 	
	b:SetNormalTexture(texture)
	b:SetHeight(buttonSize)
	b:SetWidth(buttonSize)
	b:Show()
	b:RegisterForClicks("AnyUp")
	AttachTooltips(b, info)

	local cd = nil
	if includecooldown then
		--only adding the RequestAction script to buttons which have a cooldown - the combat window buttons
		b:SetScript("OnClick", RequestAction)
		--

		cd = CreateFrame("Cooldown",nil,b)
		cd:SetAllPoints()
		cd:SetScript("OnHide", function(self)
			b:SetAlpha(opacityAvailable);
			self:SetAlpha(opacityAvailable);
		end)
		cd:SetScript("OnShow", function(self)
			b:SetAlpha(opacityCD);
			self:SetAlpha(opacityCD);
		end)
	end
	return b, cd, f
end

function Telepathy:CreateMyAbilityButton(info)
	
	local spec = GetCurrentActiveSpecName()
	if PublishedAbilities[spec] and PublishedAbilities[spec][info.id] and PublishedAbilities[spec][info.id].button then
		return PublishedAbilities[spec][info.id].button
	end
	
	local buttonSize = db.ButtonSizeLocal
	local start, duration, enabled
	local mp = self.MainPanel
	local b = CreateFrame("Button", nil, mp, "SecureActionButtonTemplate")
	b:SetScript("OnUpdate", OnCDButtonUpdate)
	b:SetAttribute("type", info.type)	
	b:SetAttribute("id", info.id)
	--b:SetFrameLevel(mp:GetFrameLevel()+10)
	b.id = info.id
	b.info = info
	if (info.type == "spell") then				
		b:SetAttribute("spell", info.name)
		start, duration, enabled = GetSpellCooldown(info.id)
	end
	if (info.type == "item") then
		b:SetAttribute("item", info.name)
		start, duration, enabled = GetItemCooldown(info.id)
	end
	b:ClearAllPoints()
	b:SetPoint("TOPLEFT",mp,"TOPLEFT",0,0)
	local texture
	if info.type == "spell" then
		local _, _, spellicon = GetSpellInfo(info.id)	
		texture = spellicon
	elseif info.type == "item" then
		local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(info.id)
		texture = itemicon
	end   	 	
	b:SetNormalTexture(texture)
	b:SetHeight(buttonSize)
	b:SetWidth(buttonSize)
	b:Show()

	local cd = CreateFrame("Cooldown",nil,b)
	cd:SetAllPoints()
	cd:SetCooldown(start, duration)			
	info.button = b
	info.cd = cd	
	
	AttachTooltips(b, info)

	local highlight = b:CreateTexture()
	highlight:SetHeight(b:GetHeight() + (buttonPadding*2))
	highlight:SetWidth(b:GetWidth() + (buttonPadding*2))
	highlight:ClearAllPoints(b)
	highlight:SetPoint("CENTER")
	highlight:SetTexture(0,0,0,0)
	info.highlight = highlight

	--add to main panel?
	tinsert(ButtonCache.Local, b)
	mp:Refresh()
end

function GetCooldownRemaining(id)
	local started, duration, a = GetSpellCooldown(id)
	return (started + duration - GetTime()), started, duration
end

function Telepathy:PublishAbility(specname, abilityname, bRemove, bDeferUpdate)
	self:Trace("PublishAbility("..specname..","..abilityname..")")
	
	if ( not PublishedAbilities[specname] ) then
		PublishedAbilities[specname] = {}
	end


	local Id
	if type(abilityname) == "string" then
		local _, _, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, Name = string.find(abilityname, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")	
		if Id then
			self:Trace(Id, ":", Name)
			abilityname = Id	
		end
	else
		Id = abilityname
	end

	local atype
	local info = GetSpellInfo(Id)
	if info then atype = "spell" else atype = "item" end
	if not info then info = GetItemInfo(Id) end
	if not info then return end -- dunno what this is
	local id = tonumber(Id)

	--what are we publishing?
	if not bRemove then		
		local p = PublishedAbilities[specname][id] or {}
		PublishedAbilities[specname][id] = p
	
		p.type = atype
		p.name = info
		p.id = id
		p.requestable = true

		self:Trace(p.name, p.type, p.id)

		local useable, mana = IsUsableSpell(abilityname)
		self:Trace("useable :", useable, " lowmana :", mana)

		if ( specname == GetCurrentActiveSpecName() ) then
			self:CreateMyAbilityButton(p)
		end
	else
		local info = PublishedAbilities[specname][Id]
		if info.button then
			info.button:Hide()
		end
		removeientry(ButtonCache.Local, info.button)
		PublishedAbilities[specname][Id] = nil
		self.MainPanel:Refresh()
	end

	if not bDeferUpdate then
		local cd_remaining, _, duration = GetCooldownRemaining(id)
		local pubstring = strconcat(id,"|",atype:sub(1,1),"|","y","|",round(cd_remaining,2),"|",duration)
		if bRemove then
			pubstring = pubstring.."|-"
		end	
		self:ComPublish(pubstring, currentChannel, nil)
	end
end

function Telepathy:RecreateLocalButtonCache()
	
	-- clear all the buttons from all the abilities if they exist
	for spec, spectab in pairs(PublishedAbilities) do
		for Id,v in pairs(spectab) do
			if v.button then
				v.button:Hide()
			end
			removeientry(ButtonCache.Local, v.button)
		end
	end
	
	-- add the current ability buttons
	local spec = GetCurrentActiveSpecName()
	local abilities = PublishedAbilities[spec]
	
	if ( abilities ) then
		for k,v in pairs(abilities) do
			self:CreateMyAbilityButton(v)
		end
	end

	self.MainPanel:Refresh()

end

function Telepathy:UNIT_SPELLCAST_SUCCEEDED(event, ...)
	local unit, _, _, _, spellid = ...
	if UnitIsPlayer(unit) then
		local spec = GetCurrentActiveSpecName()
		if ( PublishedAbilities[spec] ) then			
			local pubinfo = PublishedAbilities[spec][spellid]
			if pubinfo and pubinfo.button then
				pubinfo.button.needsupdate = true
				if pubinfo.highlight then
					pubinfo.highlight:Hide()				
				end
				self:Trace("set pubinfo.button.needsupdate = true")
			end
			--also throw away any requests for this
			local request_table = ButtonCache.RequestFrames[spellid]
			if request_table then
				for k,v in ipairs(request_table) do
					self:RemoveRequester(v)
				end
			end
		end
	end
end

function Telepathy:GROUP_ROSTER_UPDATE(event,...)
	if not self.PartyChanging then
		self.PartyChanging = true
		self:Trace("GROUP_ROSTER_UPDATE: starting update timer")
		self:ScheduleTimer("OnPartyChangedTimer", 1, nil)
	end
end

function Telepathy:ACTIVE_TALENT_GROUP_CHANGED(event,...)
	local c = GetCurrentChannel()
	local newTalentGroup = ...
	self:Trace("ACTIVE_TALENT_GROUP_CHANGED, new talent spec", newTalentGroup, GetCurrentActiveSpecName(newTalentGroup))
	self:ComRefresh(c, GetCurrentActiveSpecName(newTalentGroup))
	self:RecreateLocalButtonCache()
end

function Telepathy:PLAYER_ALIVE(event, ...)
	--player is alive again after death
	--is reincarnation a published ability? if so - publish it's new cooldown
	self:Trace("Player is alive!")
	if PublishedAbilities[20608] then
		self:Trace("Player is alive!")
		--lets fake a spell cast here...
		self:UNIT_SPELLCAST_SUCCEEDED(event, "player", nil, nil, nil, 20608)
	end
end

function Telepathy:OnPartyChangedTimer()
	self.PartyChanging = false	
	local c, playercount = GetCurrentChannel()
	c = c:lower() --lower channel
	if c == "battleground" then c = "raid" end
	
	self:Trace("OnPartyChangedTimer: new channel=", c, "with", playercount, "players")

	--loop players
	local existing = {}
	local name, realm, fullname
	for i=1,playercount do
		--name, realm = UnitName(c..i)
		fullname = GetUnitName(c..i)
		if fullname then			
			--fullname = name.."-"..realm
			existing[fullname] = true
			if LinkedMembers[fullname] then
				--I have met you already
			else
				--I don't know you, please share your published abilities?
				self:Trace("Sending refresh request to new person:", fullname)
				self:ComRequest("r", 0, 0, fullname) 
			end
		end
	end
	--remove anyone who we don't have anymore
	for k,v in pairs(LinkedMembers) do
		if not existing[k] then
			local a = v.Abilities
			for k1, info in pairs(a) do
				info.Hidden = true				
			end
			self:RefreshForeignPanel(v)
			LinkedMembers[k] = nil	
			--remove from subsribed as well?
			--at the moment I will save every player you have ever played with
			--maybe at some point I will make it an option to ignore players or only guildies or something
			--
			--db.Subscribed[k] = nil
		end
	end
		
	if Telepathy_SpellBook then
		Telepathy_SpellBook.RepopulateForeignButtons()
	end
	
	--check solo-ness
	self:CheckHideSolo()
end


function Telepathy:HandleSlashCommand(args, info)
	if args == "?" or args == L["help"] then
		self:Print(L["Usage:"])
		self:Print(L["/tele or /telepathy -- shows the spell book"])
		self:Print(L["/tele |cffCED000show|r -- shows the main panel"])
		self:Print(L["/tele |cffCED000hide|r -- hides the main panel"])
		self:Print(L["/tele |cffCED000reset|r -- reset the panel positions"])		
	else
		args = args:trim()
		if args == L["reset"] then
			--reset the position of the windows
			db.MainFrameX = defaults.char.MainFrameX
			db.MainFrameY = defaults.char.MainFrameY
			db.MainFrameWidth = defaults.char.MainFrameWidth
			db.MainFrameHeight = defaults.char.MainFrameHeight						
			db.LocalX = 0
			db.LocalY = 0
			Telepathy:CreateCombatFrames()
			Telepathy:Print(L["Combat window has been reset - please reload your UI to complete the reset."])
		elseif args == "debug" then
			--toggle debugging
			teleDebug = not teleDebug
			if teleDebug then
				self:Print("Debugging is ON")
			else
				self:Print("Debugging is OFF")
			end
		elseif args == L["hide"] then
			self.MainFrame:Hide()
			self:Print(L["Telepathy frames are hidden."])
		elseif args == L["show"] then
			self.MainFrame:Show()
			self:Print(L["Telepathy frames are visible."])
		else		
			Telepathy:CreateSpellBook()
		end
	end
end

function Telepathy:HandleSlashCommandTest(args, info)
	--[[
	args = args:trim()
	Telepathy:Print(args)
	local s = tostring(args)
	local id, type, r, x = strsplit("|", s)
	Telepathy:Print(id, type, r, x)
	Telepathy:SendCommMessage(comPublish, s, "WHISPER", UnitName("player"))
	]]
	Telepathy:ComRefresh("WHISPER", UnitName("player"))
end

function Telepathy:HandleSlashCommandLoadTest(args, info)
	local sender = "Player"..random(1, 10)
	for i=20, random(22,25) do
		local spell, id = GetSpellBookItemInfo(i, "spell")
		if spell then
			local fakemessage = id.."|s|y|0|120"
			Telepathy:OnComPublish(fakemessage, "WHISPER", sender)
		end
	end

	--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
	Telepathy:RefreshForeignPanel(LinkedMembers[sender])

	sender = UnitName("player")
	for i=20, random(22,27) do
		local spell, id = GetSpellBookItemInfo(i, "spell")
		if spell then
			local fakemessage = id.."|s|y|0|120"
			Telepathy:OnComPublish(fakemessage, "WHISPER", sender)
		end
	end

	--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
	Telepathy:RefreshForeignPanel(LinkedMembers[sender])
end

function Telepathy:CreateSlashCommands()
	self:RegisterChatCommand("tele", "HandleSlashCommand")
	self:RegisterChatCommand("telepathy", "HandleSlashCommand")
	self:RegisterChatCommand("teletest", "HandleSlashCommandTest")
	self:RegisterChatCommand("teleloadtest", "HandleSlashCommandLoadTest")	
end

function Telepathy:RefreshSpellCooldown(info)
	if info.button and info.cd then		
		self:Trace("Setting spell cooldown:", info.cd_start, info.cd_duration)
		info.cd:SetCooldown(info.cd_start, info.cd_duration)
	end
end

function Telepathy:ComNotifyCD(id, type, channel, to)
	local start, duration, enabled
	if type == "spell" then
		start, duration, enabled = GetSpellCooldown(id)
	elseif type == "item" then
		start, duration, enabled = GetItemCooldown(id)
	end	
	local realstart = round(GetTime() - start, 3) --the number of seconds ago it started
	local commsg = strconcat(id,"|",type,"|",realstart,"|",duration)
	local spec = GetCurrentActiveSpecName()
	local info = PublishedAbilities[spec][id]
	info.cd_duration = duration
	info.cd_start = start
	self:SendCommMessage(comCD, commsg, channel, to)
end

function Telepathy:OnComNotifyCD(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Got Telepathy:OnComNotifyCD message: ", msg,", dist=",channel,", sender=", sender)
	
	local member = LinkedMembers[sender]	
	if not member then
		if teleDebug then
			Telepathy:Trace(sender,"has no table in LinkedMembers")
		end
		return
	end
	local tid, type, start, duration = strsplit("|", msg)
	local id = tonumber(tid)
	
	local info = member.Abilities[id]
	if not info then
		if teleDebug then
			Telepathy:Trace(sender,"has no info for id", id)
		end		
		return
	end

	Telepathy:Trace("Got foreign info for "..sender)
	local t = GetTime() - tonumber(start)
	info.cd_start = t
	info.cd_duration = tonumber(duration)
	Telepathy:RefreshSpellCooldown(info)
end

function Telepathy:ComPublish(msg, channel)
	Telepathy:SendCommMessage(comPublish, msg, channel, UnitName("player"))
	--Telepathy:SendCommMessage(comPublish, msg, "WHISPER", UnitName("player"))
end

function Telepathy:OnComPublish(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Got Telepathy:OnComPublish message: ", msg,", dist=",channel,", sender=", sender)
	-- in the format: [spellid/itemid]|[s/i]|[y/n]
	local id, type, r, cd_remain, duration, delete = strsplit("|", msg)
	if type == "s" then
		type = "spell"
	elseif type == "i" then
		type = "item"
	end
	local requestable = (r == "y")

	if delete == "-" then
		--removing this abilitiy
		local mt = LinkedMembers[sender]
		local id = tonumber(id)
		if mt and mt.Abilities then
			mt.Abilities[id].Hidden = true
			--Telepathy.ForeignPanel:MarkTableForUpdate(mt)
			Telepathy:RefreshForeignPanel(mt)
			mt.Abilities[id] = nil
			if Telepathy_SpellBook then
				Telepathy_SpellBook.RepopulateForeignButtons()
			end
		end
	else
		Telepathy:RegisterForeignAbility(tonumber(id), type, requestable, sender, tonumber(cd_remain), tonumber(duration))
	end
end

-- send out a complete table of my abilities
function Telepathy:ComRefresh(channel, to, specificSpec)
	local specname = specificSpec or GetCurrentActiveSpecName()
	local publishedSpecAbilities = PublishedAbilities[specname]
	local sendtable = {}
	self:Trace("Preparing ComRefresh for spec "..specname)
	if ( publishedSpecAbilities ) then
		for k,v in pairs(publishedSpecAbilities) do
			local remain, _, duration = GetCooldownRemaining(v.id)
			local s = { id = v.id, r = v.requestable, t = v.type, cd = remain, d = duration }
			tinsert(sendtable, s)
		end
	else
		self:Trace("publishedSpecAbilities table for "..specname.." is nil")
	end
	local msg = self:Serialize(sendtable)
	Telepathy:Trace("Sending ComRefresh:", msg, "to", to, "via", channel)
	Telepathy:SendCommMessage(comRefresh, msg, channel, to)
end

function Telepathy:OnComRefresh(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Received ComRefresh from", sender, ":", msg)

	-- the  refresh message is a table with all published ability info in the form
	-- { id, type, requestable }
	local ok, data = Telepathy:Deserialize(msg)
	if ok and type(data) == "table" then
		local addedabilities = {}
		for k,v in pairs(data) do			
			Telepathy:Trace("ComRefresh: Registering foreing ability for", sender)
			Telepathy:RegisterForeignAbility(v.id, v.t, v.r, sender, v.cd, v.d, true)
			addedabilities[v.id] = true
		end

		if LinkedMembers[sender] then
			local mt_abilities = LinkedMembers[sender].Abilities

			-- remove any abilties that aren't there anymore
			for k,v in pairs(mt_abilities) do
				if not addedabilities[k] then
					Telepathy:Trace("Removing foreing ability from", sender, k)
					if v.container then 
						removeientry(ButtonCache.Foreign, v.container)
						v.container:Hide() 
					end
					mt_abilities[k] = nil
				end
			end
		end

		if Telepathy_SpellBook then
			Telepathy_SpellBook.RepopulateForeignButtons()
		end
		--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
		Telepathy:RefreshForeignPanel(LinkedMembers[sender])
	end
end


--request types are "r [refresh] or a [ability]"
function Telepathy:ComRequest(requesttype, id, type, who)
	local msg = strconcat(requesttype,"|",id,"|",type)
	
	--not sending messages to offline players, cos that is just stupid
	if UnitExists(who) and UnitIsConnected(who) then		
		Telepathy:Trace("Sending request to",who,":",msg)
		Telepathy:SendCommMessage(comRequest, msg, "WHISPER", who)	
	else
		Telepathy:Trace("Not sending message to",who,"who is offline.")
	end
end

function Telepathy:RequestBigRefresh()	
	local msg = "r|.|"
	Telepathy:Trace("Sending big refresh request message to", currentChannel)
	Telepathy:SendCommMessage(comRequest, msg, currentChannel, nil)	
end

function Telepathy:OnComRequest(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	local req, tid, type = strsplit("|", msg)
	Telepathy:Trace("OnComRequest(",req,id,type,")")

	if req == "a" then
		local id = tonumber(tid)
		local spellname = GetSpellInfo(id)
	
		Telepathy:Trace("Got ability request for",spellname)

		--ask for an ability to be cast, highlight the button
		local spec = GetCurrentActiveSpecName()
		local abilities = PublishedAbilities[spec]
		local info = abilities[id]
		if info and info.button and info.highlight then
			local h = info.highlight
			Telepathy:Trace("Highlighting button")
			h:SetTexture(1, 1, 0, 0.5)
			h:Show()
			h.HideTime = GetTime() + 5
			if db.RequestSound and string.len(db.RequestSound) then
				PlaySoundFile(sounds[db.RequestSound])
			end
			Telepathy:AddRequester(id, sender)
		else
			Telepathy:Trace("Not highlighting button")
			if info and not info.button then
				Telepathy:Trace("no button!")
			elseif info and not info.highlight then
				Telepathy:Trace("no highlight!")
			elseif not info then
				Telepathy:Trace("no info!")
			end
		end
	elseif req == "r" then
		--player is asking for my list of abilities
		Telepathy:Trace("Responding to refresh request from", sender,"...")
		Telepathy:ComRefresh("WHISPER", sender)
	end
end

function Telepathy:OnCooldownHighlightCheckTimer()
	local now = GetTime()
	local activeSpec = GetCurrentActiveSpecName()
	if ( PublishedAbilities[activeSpec] ) then
		for k,info in pairs(PublishedAbilities[activeSpec]) do	
			local h = info.highlight
			if h and h:IsVisible() and h.HideTime then
				if now > h.HideTime then
					Telepathy:Trace("Hiding highlight texture")
					h:Hide()
					h.HideTime = 0
				end
			elseif not h then
				Telepathy:Trace("highlight for", v.Name, "spell id", info.id, "is nil")
			end	
		end
	end
end


function Telepathy:CreateCombatFrames()
	
	if self.MainFrame then
		self.MainFrame:Hide()
		self.MainPanel:Hide()
		self.ForeignPanel:Hide()
		self.MainFrame = nil
		self.MainPanel = nil
		self.ForeignPanel = nil
	end

	self:CreateMainFrame()
	self:CreateMainPanel()
	self:CreateForeignPanel();

	self:InitialResizeWorkaround()
end

function Telepathy:InitialResizeWorkaround()
	--lets just do a quick "resize" of this window, to get everything squared away nicely
	local mp, mf = self.MainPanel, self.MainFrame
	self.OnMainFrameSizeChanged(mf, mf:GetWidth(), mf:GetHeight())
	mp:Refresh()
	self.OnMainFrameSizeChanged(mf, mf:GetWidth(), mf:GetHeight())
end

function Telepathy:OnEnable()
    --set up the config options
	self.db = LibStub("AceDB-3.0"):New("TelepathyDB", defaults, true)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Telepathy", options, {"telepathy"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Telepathy", "Telepathy")

	db = self.db.char
	
	--load local values
	buttonPadding = db.ButtonPadding
	opacityCD = db.OpacityCD
	opacityAvailable = db.OpacityAvailable
	
	--create frames
	self:CreateCombatFrames()

	--register slash commands
	self:CreateSlashCommands()

	--register events
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("GROUP_ROSTER_UPDATE")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	local _, classname = UnitClass("player")
	if classname == "SHAMAN" then
		--we must track deaths, in case the shaman ankhs
		self:Trace("Player is a shaman, so tracking the PLAYER_ALIVE event")
		self:RegisterEvent("PLAYER_ALIVE")
	end

	--register communication functions
	self:RegisterComm(comPublish, self.OnComPublish)
	self:RegisterComm(comCD, self.OnComNotifyCD)
	self:RegisterComm(comRequest, self.OnComRequest)
	self:RegisterComm(comRefresh, self.OnComRefresh)

	-- upgrade... remove any numeric keys from the published table
	for k,v in pairs(db.Published) do
		if type(k) == "number" then
			db.Published[k] = nil		
		end
	end

	--load the previously published abilities, saved in the db.Published table
	for spec, spectable in pairs(db.Published) do		
		--ensure a table
		if type(spectable) == "table" then
			for k,v in pairs(spectable) do
				if v.id then
					self:PublishAbility(spec, v.id, false, true)
				end
			end
		end
	end

	db.Published = PublishedAbilities

	local c = GetCurrentChannel()
	self:ComRefresh(c)

	--start cooldown highlight timer
	self:ScheduleRepeatingTimer("OnCooldownHighlightCheckTimer", 2, nil)

	local welcome = function()
		--welcome message
		local v = GetAddOnMetadata("Telepathy", "Version")
		self:Print(string.format(L["|cff20D6E6Welcome to Telepathy %s!|r"], v))
		self:Print(L["|cff20D6E6Any feedback is welcome on the Telepathy page at Curse.com, and anyone able to translate for other languages please visit WowAce.com|r"])
		self:Print(L["|cff20D6E6Open the Telepathy spellbook by typing|r |cffD9DC2A/telepathy|r"])
		self:Print(L["|cff20D6E6Get help by typing|r |cffD9DC2A/telepathy help|r"])	
	end

	self:ScheduleTimer(welcome, 3)
	
	--check solo-ness
	self:CheckHideSolo()

	--
	self:InitialResizeWorkaround()
end

function Telepathy:OnDisable()
    -- Called when the addon is disabled
end
