﻿-- Author      : Arclight
-- Create Date : 2/19/2012 11:09:13 AM

local pairs = pairs
local ipairs = ipairs
local max = max
local min = min
local WrapPanel = {}

function CreateWrapPanel(parent)
	local p = CreateFrame("Frame", nil, parent)
	p:SetAllPoints()
	
	--setup functions
	p.Refresh=WrapPanel.Refresh
	p.SetContext=WrapPanel.SetContext
	p.SetPadding=WrapPanel.SetPadding
	p.SetMargins=WrapPanel.SetMargins
	p.PreferredWidth=WrapPanel.PreferredWidth
	--setup properties
	p.Padding = {0, 0}
	p.Margins = {0, 0, 0, 0}
	p.IsWrapPanel = true
	p.AutoSizeY = false -- allows the wrap panel to grow/shrink vertically if there is no horizontal space
	p.AutoSizeX = false -- allows the wrap panel to grow/shrink horizontally depending on vertical space
	p.MinHeight = 0
	p.MinWidth = 0

	return p
end

-- recalculate this panel, wrapping the children
function WrapPanel:Refresh()
	
	if not self.Context then return end

	--print("WrapPanel:Refresh() for: ".. (self.Tag or "Unknown"))
	
	local padding = self.Padding
	local margins = self.Margins
	local w, h = 0, 0
	local maxw, maxh = self:GetWidth() - (margins[1] + margins[3]), self:GetHeight() - (margins[2] + margins[4])
	local currenth = self:GetHeight()
	local x, y, nexty, highestx = margins[1], margins[2] * -1, margins[2] * -1, 0

	local bSomethingHidden = false
	for i, p in ipairs(self.Context) do
		w, h = p:GetWidth(), p:GetHeight()
		--if this child is also a wrap panel, can we shrink it to fit in the available width?
		if p.IsWrapPanel and p.AutoSizeX and p.MinWidth ~= 0 and p.MinWidth < maxw then								
				w = maxw
				p:SetWidth(w)
				p:Refresh()
				h = p:GetHeight()
				w = p:GetWidth()
		end
		if x ~= margins[1] and (x + w > maxw or (p.WrapHint and p.WrapHint == "NEWLINE")) then
			--wrap
			y = nexty
			x = margins[1]
		end
					
		p:ClearAllPoints()
		p:SetPoint("TOPLEFT", self, "TOPLEFT", x, y - (p.WrapYOffset or 0))
		highestx = max(highestx, x + w)

		x = x + w + padding[1]
		nexty = min(nexty, y - h - padding[2])

		if abs(y - h + margins[2]) > maxh then
			p.WrapPanelHidden = true
			p:Hide()
			bSomethingHidden = true
		elseif p.WrapPanelHidden then
			p.WrapPanelHidden = false
			p:Show()			
		end

		nexty = min(nexty, y - h - padding[2])
	end

	if self.AutoSizeY then
		local newreqsize = (nexty*-1)+margins[4]
		newreqsize = max(newreqsize, self.MinHeight or newreqsize)
		self:SetHeight(newreqsize)		
	end
	if self.AutoSizeX then
		local newreqsize = highestx+margins[3]
		newreqsize = max(newreqsize, self.MinWidth or newreqsize)
		self:SetWidth(newreqsize)
	end

	if bSomethingHidden then
		if self.MoreLabel then
			self.MoreLabel:Show()
		else
			local extralabel = self:CreateFontString(nil,"ARTWORK")
			extralabel:SetFont("Fonts\\ARIALN.TTF",18,"OUTLINE") --(GameFontNormal:GetFont())
			extralabel:SetText("...")
			extralabel:SetPoint("BOTTOM", 0, 5)			
			extralabel:Show()
			self.MoreLabel = extralabel
		end
	elseif self.MoreLabel then
		self.MoreLabel:Hide()
	end
end

-- calculate the width of panel, assuming it can fit in one whole row
function WrapPanel:PreferredWidth()
	if not self.Context then return self.Margins[1] + self.Margins[3] end

	local xpadding = self.Padding[1]	
	local x = self.Margins[1] + self.Margins[3]

	for i, p in ipairs(self.Context) do
		w = p:GetWidth()
		x = x + w + xpadding		
	end
	return x
end

function WrapPanel:SetPadding(x, y)
	self.Padding[1] = x
	self.Padding[2] = y or x
	self:Refresh()
end

function WrapPanel:SetMargins(x1, y1, x2, y2)
	self.Margins = {x1, y1 or x1, x2 or x1, y2 or y1 or x1}
	self:Refresh()
end

function WrapPanel:SetContext(context, suppressrefresh)
	self.Context = context
	if not suppressrefresh then self:Refresh() end
end